/*
 * main.c
 *
 * Copyright (c) 2002 Maksim Yevmenkin <m_evmenkin@yahoo.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main.c,v 1.12 2007/01/30 00:35:23 max Exp $
 * $FreeBSD$
 */

#include <bluetooth.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <libgen.h>
#include <limits.h>
#include <obex.h>
#include <sdp.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "obexapp.h"
#include "client.h"
#include "event.h"
#include "log.h"
#include "server.h"
#include "transport.h"
#include "util.h"

int rfcomm_channel_lookup(bdaddr_t const *local, bdaddr_t const *remote,
			int service, int *channel, int *error);

static void usage(char const *prog);

/*
 * Main
 */

int
main(int argc, char *argv[])
{
	struct sigaction	sa;
	char			*ep = NULL, *pri_name = NULL;
	int			n, service, noninteractive;
	context_t		context;
	obex_ctrans_t		custfunc;

	/* Initialize locale */
	if (obexapp_util_locale_init() < 0)
		errx(1, "Could not initialize locale");

	/* Ignore SIGCHLD */
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = SIG_IGN;
	sa.sa_flags = SA_NOCLDWAIT;
	if (sigaction(SIGCHLD, &sa, NULL) < 0)
		err(1, "Could not sigaction(SIGCHLD)");

	/* Prepare context */
	memset(&context, 0, sizeof(context));
	context.addr.rfcomm_len = sizeof(context.addr.rfcomm_len);
	context.addr.rfcomm_family = AF_BLUETOOTH;
	context.tfd = context.sfd = -1;
	context.detach = 1;

	context.ls_size = OBEXAPP_BUFFER_SIZE;
	if ((context.ls = (char *) malloc(context.ls_size)) == NULL)
		errx(1, "Could not allocate context.ls");

	if ((context.root = (char *) malloc(PATH_MAX)) == NULL)
		errx(1, "Could not allocate context.root");
	context.root[0] = '\0';

	if ((context.file = (char *) malloc(PATH_MAX)) == NULL)
		errx(1, "Could not allocate context.file");

	if ((context.temp = (char *) malloc(PATH_MAX)) == NULL)
		errx(1, "Could not allocate context.temp");

	context.tbuffer = (char *) malloc(OBEXAPP_BUFFER_SIZE);
	if (context.tbuffer == NULL)
		errx(1, "Could not allocate context.tbuffer");

	context.sbuffer = (char *) malloc(OBEXAPP_BUFFER_SIZE);
	if (context.sbuffer == NULL)
		errx(1, "Could not allocate context.sbuffer");

	context.mtu = OBEX_MAXIMUM_MTU;

	memset(&custfunc, 0, sizeof(custfunc));
	custfunc.connect = obexapp_transport_connect;
	custfunc.disconnect = obexapp_transport_disconnect;
	custfunc.listen = obexapp_transport_listen;
	custfunc.write = obexapp_transport_write;
	custfunc.handleinput = obexapp_transport_handle_input;
	custfunc.customdata = &context;

	/* Process command line options */
	service = noninteractive = 0;
	while ((n = getopt(argc, argv, "a:A:cC:dDfhl:m:nr:Ssu:")) != -1) {
		switch (n) {
		case 'a':
			if (!bt_aton(optarg, &context.addr.rfcomm_bdaddr)) {
				struct hostent	*he = bt_gethostbyname(optarg);

				if (he == NULL)
					errx(1, "%s: %s", optarg,
						hstrerror(h_errno));

				memcpy(&context.addr.rfcomm_bdaddr, he->h_addr,
					sizeof(context.addr.rfcomm_bdaddr));
			}
			break;

		case 'A':
			if (!bt_aton(optarg, &context.saddr)) {
				struct hostent	*he = bt_gethostbyname(optarg);

				if (he == NULL)
					errx(1, "%s: %s", optarg,
						hstrerror(h_errno));

				memcpy(&context.saddr, he->h_addr,
						sizeof(context.saddr));
			}
			break;

		case 'c': /* client */
			context.server = 0;
			break;

		case 'C': /* channel to listen on or connect to */
			context.addr.rfcomm_channel = strtoul(optarg, &ep, 10);
			if (*ep != '\0') {
				context.addr.rfcomm_channel = 0;

				switch (tolower(optarg[0])) {
				case 'i': /* IrMC */
					service = SDP_SERVICE_CLASS_IR_MC_SYNC;
					break;

				case 'f': /* OBEX File Transfer */
					service = SDP_SERVICE_CLASS_OBEX_FILE_TRANSFER;
					break;

				case 'o': /* OBEX Push */
					service = SDP_SERVICE_CLASS_OBEX_OBJECT_PUSH;
					break;
				}
			}
			break;

		case 'd': /* do not detach server */
			context.detach = 0;
			break;

		case 'D': /* use stdin/stdout */
			context.direct = 1;
			break;

		case 'f': /* connect to Folder Browsing Service */
			context.fbs = 1;
			break;

		case 'l': /* log priority */
			pri_name = optarg;
			break;

		case 'm': /* OBEX MTU */
			context.mtu = strtoul(optarg, &ep, 10);

			if (*ep != '\0' ||
			    context.mtu < OBEX_MINIMUM_MTU ||
			    context.mtu > OBEX_MAXIMUM_MTU)
				errx(1, "Invalid OBEX MTU: %s", optarg);
			break;

		case 'n': /* non-interactive client mode */
			if (context.server)
				usage(basename(argv[0]));
				/* NOT REACHED */

			noninteractive = 1;
			break;

		case 'r': /* root */
			if (realpath(optarg, context.root) == NULL)
				err(1, "Could not realpath(%s)", optarg);
			break;

		case 's': /* server */
			if (noninteractive)
				usage(basename(argv[0]));
				/* NOT REACHED */

			context.server = 1;
			break;

		case 'S': /* secure */
			context.secure = 1;
			break;

		case 'u': /* user */
			context.user = optarg;
			break;

		case 'h':
		default:
			usage(basename(argv[0]));
			/* NOT REACHED */
		}
	}

	argc -= optind;
	argv += optind;

	if (!context.server) {
		if (memcmp(&context.addr.rfcomm_bdaddr,
		    NG_HCI_BDADDR_ANY, sizeof(context.addr.rfcomm_bdaddr)) == 0)
			errx(1, "Must specify server BD_ADDR");

		/* Check channel, if was not set then obtain it via SDP */
		if (context.addr.rfcomm_channel == 0 && service != 0) {
			int	channel;

			if (rfcomm_channel_lookup(NULL,
						&context.addr.rfcomm_bdaddr,
						service, &channel, &n) != 0)
				errc(1, n, "Could not obtain RFCOMM channel");

			context.addr.rfcomm_channel = (u_int8_t) channel;
		}
	}

	if (context.addr.rfcomm_channel <= 0 ||
	    context.addr.rfcomm_channel > 30)
		errx(1, "Invalid RFCOMM channel %d",
			context.addr.rfcomm_channel);

	log_open("obexapp", pri_name, 0);

	/* Detach server (if required) */
	if (context.server && context.detach) {
		pid_t	pid = fork();

		if (pid == (pid_t) -1) {
			log_err("%s(): Could not fork. %s (%d)",
				__func__, strerror(errno), errno);
			exit(1);
		}

		if (pid != 0)
			exit(0);

		if (daemon(0, 0) < 0) {
			log_err("%s(): Could not daemon. %s (%d)",
				__func__, strerror(errno), errno);
			exit(1);
		}
	}

	/* Initialize OBEX */
	context.handle = OBEX_Init(OBEX_TRANS_CUST, obexapp_event_handler, 0);
	if (context.handle == NULL) {
		log_err("%s(): OBEX_Init failed", __func__);
		exit(1);
	}

	OBEX_SetUserData(context.handle, &context);

	/* Register custom transport */
	if (OBEX_RegisterCTransport(context.handle, &custfunc) < 0) {
		log_err("%s(): OBEX_RegisterCTransport failed", __func__);
		exit(1);
        }

	if (context.server)
		n = obexapp_server(context.handle);
	else if (noninteractive)
		n = obexapp_non_interactive_client(context.handle, argc, argv);
	else
		n = obexapp_client(context.handle);

	OBEX_Cleanup(context.handle);

	free(context.ls);
	free(context.root);
	free(context.file);
	free(context.temp);
	free(context.tbuffer);
	free(context.sbuffer);
	memset(&context, 0, sizeof(context));

	obexapp_util_locale_fini();

	log_close();

	exit(n);
} /* main */

/*
 * Display usage and exit
 */

static void
usage(char const *prog)
{
	fprintf(stderr,
"Usage: %s options\n" \
"Options:\n" \
"-a bdaddr      bdaddr to listen on/connect to (required for client)\n" \
"-A bdaddr      local bdaddr to bind to (client mode)\n" \
"-c             act as client (default)\n" \
"-C channel     RFCOMM channel (required)\n" \
"-d             run server in foreground\n" \
"-D             use stdin/stdout instead of socket\n" \
"-f             use Folder Browsing Service\n" \
"-h             display this message\n" \
"-l priority    set least severe log priority\n" \
"-m MTU         set OBEX MTU\n" \
"-n             non-interactive mode (client mode)\n" \
"-r path        set path to root directory\n" \
"-s             act as server\n" \
"-S             secure mode (recommended)\n" \
"-u user        set user ID to 'user'\n" \
"",		prog);
	exit(255);
} /* usage */

